import click
import logging

from es_translator import EsTranslator
from tempfile import mkdtemp
# Module from the same package
from es_translator.interpreters import Apertium, Argos
from es_translator.interpreters.apertium.pairs import Pairs
from es_translator.logger import add_syslog_handler, add_stdout_handler
from es_translator.tasks import app as celery_app


def validate_loglevel(ctx, param, value):
    try:
        if isinstance(value, str):
            return getattr(logging, value)
        return int(value)
    except (AttributeError, ValueError):
        raise click.BadParameter(
            'must be a valid log level (CRITICAL, ERROR, WARNING, INFO, DEBUG or NOTSET)')


def validate_progressbar(ctx, param, value):
    # If no value given, we activate the progress bar only when the
    # stdout_loglevel value is higher than INFO (20)
    return value if value is not None else ctx.params['stdout_loglevel'] > 20


def validate_interpreter(ctx, param, value):
    interpreters = (Apertium, Argos, )
    for interpreter in interpreters:
        if value.upper() == interpreter.name.upper():
            return interpreter.name
    names = (interpreter.name for interpreter in interpreters)
    raise click.BadParameter(
        'must be a valid interpreter name (%s)' %
        ', '.join(names))


@click.command()
@click.option('-u', '--url', help='Elastichsearch URL',
              default="http://localhost:9200")
@click.option('-i', '--index', help='Elastichsearch Index', required=True)
@click.option('-r',
              '--interpreter',
              help='Interpreter to use to perform the translation',
              default='ARGOS',
              callback=validate_interpreter)
@click.option('-s',
              '--source-language',
              help='Source language to translate from',
              required=True,
              default=None)
@click.option('-t',
              '--target-language',
              help='Target language to translate to',
              required=True,
              default=None)
@click.option('--intermediary-language',
              help='An intermediary language to use when no translation is available between the source and the target. If none is provided this will be calculated automatically.')
@click.option('--source-field',
              help='Document field to translate',
              default="content")
@click.option('--target-field',
              help='Document field where the translations are stored',
              default="content_translated")
@click.option('-q', '--query-string',
              help='Search query string to filter result')
@click.option('-d',
              '--data-dir',
              help='Path to the directory where to language model will be downloaded',
              type=click.Path(exists=True,
                              dir_okay=True,
                              writable=True,
                              readable=True),
              default=mkdtemp())
@click.option('--scan-scroll',
              help='Scroll duration (set to higher value if you\'re processing a lot of documents)',
              default="5m")
@click.option('--dry-run',
              help='Don\'t save anything in Elasticsearch',
              is_flag=True,
              default=False)
@click.option('--pool-size',
              help='Number of parallel processes to start',
              default=1)
@click.option('--pool-timeout',
              help='Timeout to add a translation',
              default=60 * 30)
@click.option('--throttle',
              help='Throttle between each translation (in ms)',
              default=0)
@click.option('--syslog-address', help='Syslog address', default='localhost')
@click.option('--syslog-port', help='Syslog port', default=514)
@click.option('--syslog-facility', help='Syslog facility', default='local7')
@click.option('--stdout-loglevel',
              help='Change the default log level for stdout error handler',
              default='ERROR',
              callback=validate_loglevel)
@click.option('--progressbar/--no-progressbar',
              help='Display a progressbar',
              default=None,
              callback=validate_progressbar)
@click.option('--plan',
              help='Plan translations into a queue instead of processing them npw',
              is_flag=True,
              default=False)
@click.option('--redis-url',
              help='Redis URL (only needed when planning translation)',
              default='redis://localhost:6379')
def translate(syslog_address, syslog_port, syslog_facility, **options):
    # Configure Syslog handler
    add_syslog_handler(syslog_address, syslog_port, syslog_facility)
    add_stdout_handler(options['stdout_loglevel'])
    # Configure celery app broker url globally
    celery_app.conf.broker_url = options['redis_url']
    # We setup the translator. Etheir if the translation is done now
    # or later, we need initialize the interpreter (Argos, Apertium, ...)
    es_translator = EsTranslator(options)
    es_translator.start()
        


@click.command()
@click.option('--broker', default='redis://redis', help='Celery broker URL.')
@click.option('--concurrency', default=1, help='Celery broker URL.')
@click.option('--stdout-loglevel',
              help='Change the default log level for stdout error handler',
              default='ERROR',
              callback=validate_loglevel)
def tasks(broker, concurrency, stdout_loglevel):
    """Starts a Celery worker."""
    celery_app.conf.broker_url = broker
    argv = [
        'worker',
        '--concurrency',
        concurrency,
        '--loglevel',
        stdout_loglevel]
    celery_app.worker_main(argv)


@click.command()
@click.option('--data-dir',
              help='Path to the directory where to language model will be downloaded',
              type=click.Path(exists=True,
                              dir_okay=True,
                              writable=True,
                              readable=True),
              default=mkdtemp())
@click.option('--local', help='List pairs available locally',
              is_flag=True, default=False)
@click.option('--syslog-address', help='Syslog address', default='localhost')
@click.option('--syslog-port', help='Syslog port', default=514)
@click.option('--syslog-facility', help='Syslog facility', default='local7')
@click.option('--stdout-loglevel',
              help='Change the default log level for stdout error handler',
              default='ERROR',
              callback=validate_loglevel)
def pairs(
        data_dir,
        local,
        syslog_address,
        syslog_port,
        syslog_facility,
        **options):
    # Configure Syslog handler
    add_syslog_handler(syslog_address, syslog_port, syslog_facility)
    add_stdout_handler(options['stdout_loglevel'])
    # Only the data-dir is needed to construct the Apertium instance, then
    # we just need to print the pair
    Pairs(data_dir, local).print_pairs()


if __name__ == '__main__':
    translate()



from argostranslate import package as argospackage
from argostranslate import translate as argostranslate
from argostranslate.utils import logger as argoslogger

from ..abstract import AbstractInterpreter
from ...logger import logger


class ArgosPairNotAvailable(Exception):
    pass


class Argos(AbstractInterpreter):
    name = 'ARGOS'

    def __init__(
            self,
            source=None,
            target=None,
            intermediary=None,
            pack_dir=None):
        super().__init__(source, target)
        # Raise an exception if an intermediary language is provded
        if intermediary is not None:
            logger.warn(
                'Argos interpreter doesnt support intermediary language')
        if pack_dir is not None:
            logger.warn(
                'Argos interpreter doesnt support custom pack directory')
        # Raise an exception if the language pair is unkown
        if not self.is_pair_available and self.has_pair:
            try:
                self.download_necessary_languages()
            except ArgosPairNotAvailable:
                raise Exception('The pair %s is not available' % self.pair)
        else:
            logger.info('Existing package(s) found for pair %s' % self.pair)

    @property
    def is_pair_available(self):
        for package in argospackage.get_installed_packages():
            if package.from_code == self.source_alpha_2 and package.to_code == self.target_alpha_2:
                return True
        return False

    @property
    def local_languages(self):
        try:
            installed_languages = argostranslate.get_installed_languages()
            return [lang.code for lang in installed_languages]
        except AttributeError:
            return []

    def download_necessary_languages(self):
        argospackage.update_package_index()
        for package in argospackage.get_available_packages():
            if package.from_code == self.source_alpha_2 and package.to_code == self.target_alpha_2:
                download_path = package.download()
                return argospackage.install_from_path(download_path)
        raise ArgosPairNotAvailable

    @property
    def translation(self):
        installed_languages = argostranslate.get_installed_languages()
        source = list(
            filter(
                lambda x: x.code == self.source_alpha_2,
                installed_languages))[0]
        target = list(
            filter(
                lambda x: x.code == self.target_alpha_2,
                installed_languages))[0]
        return source.get_translation(target)

    def translate(self, input):
        return self.translation.translate(input)



# Module from the same package
from .apertium import Apertium


class Pairs:
    def __init__(self, data_dir, local=False):
        self.apertium = Apertium(pack_dir=data_dir)
        self.local = local

    @property
    def local_pairs(self):
        return self.apertium.local_pairs

    @property
    def remote_pairs(self):
        return self.apertium.remote_pairs

    @property
    def local_pairs_to_string(self):
        return '\n'.join(self.local_pairs)

    @property
    def remote_pairs_to_string(self):
        return '\n'.join(self.remote_pairs)

    def print_pairs(self):
        if self.local:
            print(self.local_pairs_to_string)
        else:
            print(self.remote_pairs_to_string)



from os.path import isdir, isfile, islink
from sh import ln, rm


def create_symlink(source, target, options='-s', force=True):
    if isdir(source) or isfile(source):
        if force and islink(target):
            rm(target)
        ln(options, source, target)



from tempfile import NamedTemporaryFile
from functools import lru_cache
from sh import apertium, ErrorReturnCode
# Module from the same package
from .repository import ApertiumRepository
from ..abstract import AbstractInterpreter
from ...alpha import to_alpha_3_pair
from ...logger import logger


class Apertium(AbstractInterpreter):
    name = 'APERTIUM'

    def __init__(
            self,
            source=None,
            target=None,
            intermediary=None,
            pack_dir=None):
        super().__init__(source, target, intermediary, pack_dir)
        # A class to download necessary pair package
        self.repository = ApertiumRepository(self.pack_dir)
        # Raise an exception if the language pair is unkown
        if not self.is_pair_available and self.has_pair:
            try:
                self.download_necessary_pairs()
            except StopIteration:
                raise Exception('The pair is not available')
        else:
            logger.info('Existing package(s) found for pair %s' % self.pair)

    @property
    def pair_package(self):
        return self.pair_to_pair_package(self.pair)

    @property
    def is_pair_available(self):
        return not self.intermediary and self.pair in self.local_pairs

    @property
    def pairs_pipeline(self):
        if self.intermediary:
            return [
                self.intermediary_source_pair,
                self.intermediary_target_pair]
        else:
            return [self.pair_alpha_3]

    @property
    def intermediary_source_pair(self):
        return '%s-%s' % (self.source_alpha_3, self.intermediary_alpha_3)

    @property
    def intermediary_source_pair_package(self):
        return self.pair_to_pair_package(self.intermediary_source_pair)

    @property
    def intermediary_target_pair(self):
        return '%s-%s' % (self.intermediary_alpha_3, self.target_alpha_3)

    @property
    def intermediary_target_pair_package(self):
        return self.pair_to_pair_package(self.intermediary_target_pair)

    @property
    def intermediary_pairs(self):
        # Find the intermediary lang only if not given
        if self.intermediary is None:
            trunk_packages = [s.split('-') for s in self.remote_pairs]
            # Build a tree of languages and their children
            packages_tree = self.lang_tree(self.source, trunk_packages)
            # Find the first path between self.source (the root) and
            # self.target in the given tree
            self.intermediary = self.first_pairs_path(
                packages_tree, self.target)[0]
        # We build the two intermediary pairs
        return [self.intermediary_source_pair, self.intermediary_target_pair]

    @property
    def local_pairs(self):
        output = apertium('-d', self.pack_dir, '-l').strip()
        return [s.strip() for s in output.split('\n')]

    @property
    @lru_cache()
    def remote_pairs(self, module='trunk'):
        packages = self.repository.pair_packages
        pairs = []
        def package_name_to_pair(n): return '-'.join(n.split('-')[-2:])
        # Extract package within these two properties
        for attr in ['Package', 'Provides']:
            for package in packages:
                for value in package.get(attr, '').split(','):
                    pair = package_name_to_pair(value.strip())
                    pairs.append(pair)
        # Remove empty values
        return [p for p in pairs if p != '']

    def pair_to_pair_package(self, pair):
        pair_inversed = '-'.join(pair.split('-')[::-1])
        combinations = [to_alpha_3_pair(pair), to_alpha_3_pair(pair_inversed)]
        try:
            return next(p for p in self.remote_pairs if p in combinations)
        except StopIteration:
            return None

    def download_necessary_pairs(self):
        logger.info('Downloading necessary package(s) for %s' % self.pair)
        if self.any_pair_variant_in_packages:
            self.download_pair()
        else:
            self.download_intermediary_pairs()

    def download_pair(self, pair=None):
        if pair is None:
            pair = self.pair_alpha_3
        else:
            pair = to_alpha_3_pair(pair)
        # All commands must be run from the pack dir
        return self.repository.install_pair_package(pair)

    @property
    def any_pair_variant_in_packages(self):
        return self.pair_alpha_3 in self.remote_pairs

    def download_intermediary_pairs(self):
        for pair in self.intermediary_pairs:
            self.download_pair(pair)

    def lang_tree(self, lang, pairs, depth=2):
        tree = dict(lang=lang, children=dict())
        for pair in pairs:
            if lang in pair and depth > 0:
                child_lang = next(l for l in pair if l != lang)
                tree["children"][child_lang] = self.lang_tree(
                    child_lang, pairs, depth - 1)
        return tree

    def first_pairs_path(self, leaf, lang):
        path = []
        for child_leaf in leaf['children'].values():
            if self.leaf_has_lang(child_leaf, lang):
                path.append(child_leaf['lang'])
                path = path + self.first_pairs_path(child_leaf, lang)
                break
        return path

    def leaf_has_lang(self, leaf, lang):
        children = leaf['children'].values()
        return lang in leaf['children'] or any(
            self.leaf_has_lang(
                child_leaf,
                lang) for child_leaf in children)

    def translate(self, input):
        for pair in self.pairs_pipeline:
            # Create a sub-process witch can receive a input
            input = self.translate_with_apertium(input, pair)
        return input

    def translate_with_apertium(self, input, pair):
        try:
            # Works with a temporary file as buffer (opened in text mode)
            with NamedTemporaryFile(mode='w+t') as temp_input_file:
                temp_input_file.writelines(input)
                temp_input_file.seek(0)
                input_translated = apertium(
                    '-ud', self.pack_dir, pair, temp_input_file.name)
        except ErrorReturnCode as e:
            raise Exception('Unable to translate this string.')
        return str(input_translated)



class TranslatedHit():
    def __init__(self, hit, source_field='content',
                 target_field='content_translated'):
        self.hit = hit
        self.source_field = source_field
        self.target_field = target_field
        # Ensure the target field is an array
        self.hit[self.target_field] = self.translations

    @property
    def source_value(self):
        return self.hit.to_dict().get(self.source_field)

    @property
    def translations(self):
        return self.hit.to_dict().get(self.target_field, [])

    @property
    def id(self):
        return self.hit.meta.id

    @property
    def routing(self):
        # Some documents don't have a routing property
        return self.hit.meta.to_dict().get('routing', None)

    @property
    def index(self):
        return self.hit.meta.index

    @property
    def doc_type(self):
        return self.hit.meta.doc_type

    @property
    def body(self):
        body = dict(doc=dict())
        body['doc'][self.target_field] = self.translations
        return body

    def save(self, client):
        client.update(
            index=self.index,
            doc_type=self.doc_type,
            id=self.id,
            routing=self.routing,
            body=self.body)

    def add_translation(self, interpreter):
        if not self.is_translated(
                interpreter.source_name,
                interpreter.target_name,
                interpreter.name):
            self.hit[self.target_field].append(dict(
                translator=interpreter.name,
                source_language=interpreter.source_name.upper(),
                target_language=interpreter.target_name.upper(),
                content=interpreter.translate(self.source_value)))

    def is_translated(self, source_language, target_language, translator):
        def same_languages(t):
            return t['source_language'] == source_language.upper() and \
                t['target_language'] == target_language.upper() and \
                t['translator'] == translator
        return any(t for t in self.translations if same_languages(t))
